#pragma once

#include "String.h"
#include "List.h"
#include "Pair.h"
#include "Optional.h"

#include <utility>
#include <functional>

template<typename K, typename T>
class Map{

public:
    struct KvPair
    {
        KvPair() = default;

        KvPair(const K& key)
            :m_key(key)
        {
        }

        KvPair(const K& key, const T& value)
            : m_key(key),
                m_value(value)
        {
        }

        KvPair(const K& key, T&& value)
            : m_key(key),
                m_value(std::move(value))
        {
        }

        bool operator==(const KvPair& other) const
        {
            return m_key == other.m_key;
        }

        const K& key() const
        {
            return m_key;
        }

        const T& value() const 
        {
            return m_value;
        }

        T& value() 
        {
            return m_value;
        }

        K m_key;
        T m_value;
    };

    Map() = default;

    void insert(const K& key, const T& value)
    {
        auto& map_entry = m_buffer[(hash(key))];
        if (!map_entry.is_set())
        {
            map_entry.init(KvPair(key, value));
        } 
        else
        {
            map_entry.m_list->insert(KvPair(key, value));
        }
    }

    void insert(const K& key, T&& value)
    {
        auto& map_entry = m_buffer[(hash(key))];
        if (!map_entry.is_set())
        {
            map_entry.init(std::move(KvPair(key, std::move(value))));
        } 
        else
        {
            map_entry.m_list->insert(std::move(KvPair(key, std::move(value))));
        }
    }

    using onItemFunc = std::function<bool(const T& key, const T& value)>;
    void forEach(onItemFunc item_func)
    {
        for(size_t idx=0; idx<BUFFER_SIZE;idx++)
        {
            const auto& map_entry = m_buffer[idx];
            auto list_func = [key=map_entry.key(), item_func](const T& item)
            {
                return item_func(key, item);
            };
            map_entry.forEach(list_func);
        }
    }

    void clear()
    {
        for(size_t idx=0; idx<BUFFER_SIZE;idx++)
        {
            m_buffer[idx].clear();
        }
    }

    bool empty() const 
    {
        for(size_t idx=0; idx<BUFFER_SIZE; idx++)
        {
            if (m_buffer[idx].is_set())
            {
                return false;
            }   
        }
        return true; 
    }

    const Optional<T>& value(const K& key) const 
    {
        const auto& map_entry = m_buffer[(hash(key))];
        if (!map_entry.is_set())
        {
            return {};
        }
        KvPair search_term(key);
        if (auto list_item = map_entry.find(search_term); list_item != nullptr)
        {
            return list_item->value();
        }
        return {};
    }

    bool has_key(const K& key) const 
    {
        const auto& map_entry = m_buffer[(hash(key))];
        if (!map_entry.is_set())
        {
            return false;
        }
        KvPair search_term(key);
        if (auto list_item = map_entry.m_list->find(search_term); list_item != nullptr)
        {
            return true;
        }
        return false;
    }

    Vector<KvPair> items() const 
    {
        Vector<KvPair> ret;
        for(size_t idx=0; idx<BUFFER_SIZE;idx++)
        {
            const auto& map_entry = m_buffer[idx];
            if (map_entry.is_set())
            {
                map_entry.m_list->flatten(ret);
            }   
        }
        return ret;
    }

    class MapIter
    {
        public:
        MapIter(const Map* container, int index = -1)
            : m_index(index), m_container(container)
        {
        }

        bool operator!=(const MapIter& other) const 
        {
            return m_index != other.m_index;
        }

        const KvPair& operator*() const 
        {
            bool found{false};
            return m_container->get_item(m_index, found);
        }

        void operator++()
        {
            m_index++;
        }

        private:
        int m_index{-1};
        const Map* m_container{nullptr};
    };

    MapIter begin() const
    {
        if (empty())
        {
            return MapIter(this, -1);
        }
        else
        {
            return MapIter(this, 0);
        }
    }

    MapIter end() const
    {
        if (empty())
        {
            return MapIter(this, -1);
        }
        else
        {
            return MapIter(this, size());
        }
    }

    MapIter find(const K& key) const 
    {
        if (empty())
        {
            return MapIter(this, -1);
        }
        size_t count{0};
        for(size_t idx=0; idx<BUFFER_SIZE; idx++)
        {
            const auto& map_entry = m_buffer[idx];
            if (map_entry.is_set())
            {
                const auto loc_index = map_entry.m_list->find_index(key);
                if (loc_index.is_set())
                {
                    return MapIter(this, count + loc_index.value());
                }
                else
                {
                    count += map_entry.m_list->size();
                }
            }       
        }
        return MapIter(this, -1);
    }

    size_t size() const
    {
        size_t count{0};
        for(size_t kdx=0; kdx<BUFFER_SIZE; kdx++)
        {
            const auto& map_entry = m_buffer[kdx];
            if (map_entry.is_set())
            {
                count++;
                auto next_item = map_entry.m_list->next();
                while(next_item)
                {   
                    next_item = next_item->next();
                    count++;
                }
            }   
        }
        return count;
    }

private:
    struct MapEntry
    {
        bool is_set() const
        {
            return bool(m_list);
        }
        void init(const KvPair& kv)
        {
            m_list = Ptr<List<KvPair> >::create(kv);
        }
        void init(KvPair&& kv)
        {
            m_list = Ptr<List<KvPair> >::create(std::move(kv));
        }
        void clear()
        {
            m_list.reset();
        }
        Ptr<List<KvPair> > m_list;
    };

    size_t hash(const K& key) const
    {
        size_t result{0};
        for(const auto c : key.data())
        {
            result += static_cast<size_t>(c);
        }
        return result / BUFFER_SIZE;
    }

    const KvPair& get_item(size_t idx, bool& found) const
    {
        static KvPair s_bad_item;

        size_t count{0};
        for(size_t kdx=0; kdx<BUFFER_SIZE; kdx++)
        {
            const auto& map_entry = m_buffer[kdx];
            if (map_entry.is_set())
            {
                if (idx == count)
                {
                    found = true;
                    return map_entry.m_list->item();
                }
                count++;
                auto next_item = map_entry.m_list->next();
                while(next_item)
                {
                    if (idx == count)
                    {
                        found = true;
                        return next_item->item();
                    }      
                    next_item = next_item->next();
                    count++;
                }
            }   
        }
        return s_bad_item;
    }

    KvPair& get_item(size_t idx, bool& found)
    {
        static KvPair s_bad_item;

        size_t count{0};
        for(size_t kdx=0; kdx<BUFFER_SIZE; kdx++)
        {
            auto& map_entry = m_buffer[kdx];
            if (map_entry.is_set())
            {
                if (idx == count)
                {
                    found = true;
                    return map_entry.m_list->item();
                }
                count++;
                auto next_item = map_entry.m_list->next();
                while(next_item)
                {
                    if (idx == count)
                    {
                        found = true;
                        return next_item->item();
                    }      
                    next_item = next_item->next();
                    count++;
                }
            }   
        }
        return s_bad_item;
    }

    static constexpr size_t BUFFER_SIZE{5000};
    MapEntry m_buffer[BUFFER_SIZE];
};