#include "XmlParser.h"

#include "Char.h"

#include "XmlDocument.h"
#include "XmlElement.h"
#include "XmlAttribute.h"

using LS = XmlParser::LineState;
using DS = XmlParser::DocumentState;

XmlParser::XmlParser()
    : mDocumentState(XmlParser::DocumentState::Await_Prolog),
      mLineState(XmlParser::LineState::Await_Tag_Open),
      mDocument(Ptr<XmlDocument>::create()),
      mWorkingElements()
{

}

void XmlParser::processLine(const String& input)
{
    for (size_t idx=0; idx<input.size(); idx++)
    {
        switch (input[idx])
        {
            case Char::LEFT_BRACKET:
                onLeftBracket();
                break;
            case Char::RIGHT_BRACKET:
                onRightBracket();
                break;
            case Char::QUESTION_MARK:
                onQuestionMark();
                break;
            case Char::FORWARD_SLASH:
                onForwardSlash();
                break;
            case Char::EQUALS:
                onEquals();
                break;
            case Char::DOUBLE_QUOTE:
                onDoubleQuote();
                break;
            default:
                onChar(input[idx]);
                break;
        }
    }
}

void XmlParser::onChar(char c)
{
    if(Char::is_alnum(c))
    {
        onAlphaNumeric(c);
    }
    else if(Char::is_space(c))
    {
        onSpace(c);
    }
    else
    {
        onNonAlphaNumeric(c);
    }
}

void XmlParser::onAlphaNumeric(char c)
{
    switch(mLineState)
    {
        case LS::Await_Tag_Open:
            onTextStart(c);
            break;
        case LS::Await_Tag_Follower:
        case LS::Await_Tag_Name:
        {
            onTagNameStart(c);
            break;
        }
        case LS::Await_Tag_Name_End:
        {
            mWorkingTagName.push_back(c);
            break;
        }
        case LS::Await_Attribute_Name:
        {
            onAttributeNameStart(c);
            break;
        }
        case LS::Await_Attribute_Name_End:
        {
            mWorkingAttributeName.push_back(c);
            break;
        }
        case LS::Await_Attribute_Value:
        {
            break;
        }
        case LS::Await_Attribute_Value_End:
        {
            mWorkingAttributeValue.push_back(c);
            break;
        }
        case LS::Await_Text_End:
        {
            mWorkingText.push_back(c);
            break;
        }
        default:
            break;
    }
}

void XmlParser::onNonAlphaNumeric(char c)
{
    switch(mLineState)
    {
        case LS::Await_Tag_Name_End:
        {
            mWorkingTagName.push_back(c);
            break;
        }
        case LS::Await_Attribute_Name_End:
        {
            mWorkingAttributeName.push_back(c);
            break;
        }
        case LS::Await_Attribute_Value_End:
        {
            mWorkingAttributeValue.push_back(c);
            break;
        }
        case LS::Await_Text_End:
        {
            mWorkingText.push_back(c);
            break;
        }
        default:
            break;
    }
}

void XmlParser::onSpace(char c)
{
    switch(mLineState)
    {
        case LS::Await_Tag_Name_End:
            onTagNameEnd();
            break;
        case LS::Await_Attribute_Name_End:
            onAttributeNameEnd();
            break;
        case LS::Await_Text_End:
            mWorkingText.push_back(c);
            break;
        case LS::Await_Attribute_Value_End:
            mWorkingAttributeValue.push_back(c);
            break;
        default:
            break;
    }
}

void XmlParser::onLeftBracket()
{
    switch(mLineState)
    {
        case LS::Await_Tag_Open:
        {
            onTagOpen();
            break;
        }
        case LS::Await_Text_End:
        {
            onTextEnd();
            onTagOpen();
            break;
        }
        default:
            onNonAlphaNumeric(Char::LEFT_BRACKET);
            break;
        }
}

void XmlParser::onRightBracket()
{
    switch(mLineState)
    {
        case LS::Await_Tag_Name_End:
        case LS::Await_Attribute_Name:
        case LS::Await_Attribute_Name_End:
        case LS::Await_Attribute_Value:
        case LS::Await_Tag_Inline_Close:
        {
            onTagClose();
            break;
        }
        default:
            onNonAlphaNumeric(Char::RIGHT_BRACKET);
            break;
        }
}

void XmlParser::onQuestionMark()
{
    if(mLineState == LS::Await_Tag_Follower)
    {
        if(mDocumentState == DS::Await_Prolog)
        {
            onStartProlog();
        }
    }
    else if(mDocumentState != DS::Build_Prolog)
    {
        onNonAlphaNumeric(Char::QUESTION_MARK);
    }
}

void XmlParser::onForwardSlash()
{
    if (mLineState == LS::Await_Tag_Follower)
    {
        if (mDocumentState == DS::Await_Element || mDocumentState == DS::Build_Element)
        {
            mDocumentState = DS::Close_Element;
        }
    }
    else if(mLineState == LS::Await_Attribute_Name)
    {
        mLineState = LS::Await_Tag_Inline_Close;
    }
    else
    {
        onNonAlphaNumeric(Char::FORWARD_SLASH);
    }
}

void XmlParser::onEquals()
{
    if(mLineState == LS::Await_Attribute_Name_End)
    {
        onAttributeNameEnd();
    }
    else
    {
        onNonAlphaNumeric(Char::EQUALS);
    }
}

void XmlParser::onDoubleQuote()
{
    if(mLineState == LS::Await_Attribute_Value)
    {
        onAttributeValueStart();
    }
    else if(mLineState == LS::Await_Attribute_Value_End)
    {
        onAttributeValueEnd();
    }
}

void XmlParser::onTagOpen()
{
    mLineState = LS::Await_Tag_Follower;
}

void XmlParser::onTagClose()
{
    if(mDocumentState == DS::Build_Prolog)
    {
        onFinishProlog();
    }
    else if(mDocumentState == DS::Build_Element)
    {
        if(mLineState == LS::Await_Tag_Name_End)
        {
            onTagNameEnd();
        }
        else if (mLineState == LS::Await_Tag_Inline_Close)
        {
            if (!mWorkingElements.empty())
            {
                mWorkingElements.pop();
            }
        }
        onElementTagEnd();
    }
    else if(mDocumentState == DS::Close_Element)
    {
        mDocumentState = DS::Await_Element;
        mLineState = LS::Await_Tag_Open;
        if (!mWorkingElements.empty())
        {
            mWorkingElements.pop();
        }
    }
}

void XmlParser::onTextStart(char c)
{
    mWorkingText += c;
    mLineState = LS::Await_Text_End;
}

void XmlParser::onTextEnd()
{
    (*mWorkingElements.top())->setText(mWorkingText);
}

void XmlParser::onElementTagEnd()
{
    mDocumentState = DS::Await_Element;
    mLineState = LS::Await_Tag_Open;
}

void XmlParser::onTagNameStart(char c)
{
    mWorkingTagName += c;
    mLineState = LS::Await_Tag_Name_End;
    if(mDocumentState != DS::Build_Prolog && mDocumentState != DS::Close_Element)
    {
        mDocumentState = DS::Build_Element;
    }
}

void XmlParser::onTagNameEnd()
{
    if(mDocumentState == DS::Build_Prolog)
    {
        mDocument->getProlog()->setTagName(mWorkingTagName);
        mLineState = LS::Await_Attribute_Name;
    }
    else if(mDocumentState == DS::Build_Element)
    {
        auto new_element = XmlElement::Create(mWorkingTagName);
        auto working_element = new_element.get();

        if (!mDocument->getRoot())
        {
            mDocument->setRoot(std::move(new_element));
        }
        else
        {
            (*mWorkingElements.top())->addChild(std::move(new_element));
        }
        mWorkingElements.push(working_element);
        mLineState = LS::Await_Attribute_Name;
    }
}

void XmlParser::onAttributeNameStart(char c)
{
    mWorkingAttributeName += c;
    mLineState = LS::Await_Attribute_Name_End;
}

void XmlParser::onAttributeNameEnd()
{
    auto attribute = XmlAttribute::Create(mWorkingAttributeName);
    if(mDocumentState == DS::Build_Prolog)
    {
        mDocument->getProlog()->addAttribute(std::move(attribute));
    }
    else if(mDocumentState == DS::Build_Element)
    {
        (*mWorkingElements.top())->addAttribute(std::move(attribute));
    }
    mLineState = LS::Await_Attribute_Value;
}

void XmlParser::onAttributeValueStart()
{
    mWorkingAttributeValue = "";
    mLineState = LS::Await_Attribute_Value_End;
}

void XmlParser::onAttributeValueEnd()
{
    if(mDocumentState == DS::Build_Prolog)
    {
        mDocument->getProlog()->getAttribute(mWorkingAttributeName)->setValue(mWorkingAttributeValue);
    }
    else if(mDocumentState == DS::Build_Element)
    {
        (*mWorkingElements.top())->getAttribute(mWorkingAttributeName)->setValue(mWorkingAttributeValue);
    }
    mLineState = LS::Await_Attribute_Name;
}

void XmlParser::onStartProlog()
{
    mDocumentState = DS::Build_Prolog;
    mLineState = LS::Await_Tag_Name_End;
}

void XmlParser::onFinishProlog()
{
    mDocument->getProlog()->update();
    mDocumentState = DS::Await_Element;
    mLineState = LS::Await_Tag_Open;
}

Ptr<XmlDocument> XmlParser::getDocument()
{
    return std::move(mDocument);
}
